/***********************************************************************/
/*                                                                     */
/*  This product contains:                                             */
/*     "Restricted Materials of IBM"                                   */
/*     (c) Copyright 1985 by the Massachusetts Institute of Technology */
/*     (c) Copyright 1986, 1995 IBM Corp.                              */
/*     All Rights Reserved                                             */
/*     Licensed Materials-Property of IBM                              */
/*     Program Number:   5765-505                                      */
/*     Program Name:     IBM AIX Print Services Facility/6000          */
/*                                                                     */
/*  THIS PROGRAM IS PROVIDED ON AN "AS-IS" BASIS AND NO                */
/*  SUPPORT OR MAINTENANCE WILL BE PROVIDED WITH RESPECT TO THIS       */
/*  PROGRAM SAMPLE.  The licensee of the PSF product is free to        */
/*  copy, revise, modify, and make derivative works of this program    */
/*  sample as they see fit.                                            */
/*  The licensee of this sample code may not resell this code or       */
/*  include it in a product that is resold.                            */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>


#define FALSE   0
#define TRUE    1

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

#define send(s,b,l,f)	write(s,b,l)
#define recv(s,b,l,f)	read(s,b,l)

#define soclose(s)	close(s)
#define tcperrno()	errno

#define CHAR	char
#define BYTE	unsigned char
#define VOID	void
#define USHORT	unsigned short
#define ULONG	unsigned long
      
#include "lprafp.h"
#if defined(LPRMON)
#include "list.h"
#include "lprmon.h"
#endif
#include "getopts.h"

/*********************/
/* Global Variables  */
/* to all modules    */
/*********************/
#ifdef QUIET
int   quietflag            = TRUE;
#else
int   quietflag            = FALSE;
#endif
int   fflag                = FALSE;
int   bflag                = FALSE;
int   nobeep               = FALSE;
char  *givenfilename       = NULL;
USHORT retries             = 3;
ULONG delay                = 10000L;
char  *server              = (char *) NULL;     /* Who we are talking to */
char  printer[20];
char  user[20];
CHAR  lpdopts[256]         = "";
char  *cp;

/*********************/
/* Global Variables  */
/* to this module    */
/*********************/
static USHORT recv_reply( int s, char *buf, unsigned len );

static char   cant_send[] = "Unable to transmit: \n";
static char   buffer[100];
static char   filename[22];      /* filename, max size=22  */
                                 /*          1         2   */
                                 /* 1234567890123456789012 */
                                 /* ---------------------- */
                                 /* cfsXXXxxx.yyy.zzz.ppp0 */
static char   tempcmds[512];
static char   cmds[512];

/*********************/
/* DEFINES           */
/*********************/
#define MAXRETRIES 9

#define COUNT_FILENAME                 "/tmp/.lprafp_count"

int get_rolling_count(void)
{
  FILE *count_fp;
  mode_t old_umask, new_umask;
  char tmp_buf[10];
  int  rolling_count;

  count_fp = fopen(COUNT_FILENAME, "r");
  if (count_fp != NULL)
  {
    fread(tmp_buf, 1, 3, count_fp);
    fclose(count_fp);
    tmp_buf[3] = 0;
    rolling_count = atoi(tmp_buf);
    rolling_count++;
    if(rolling_count > 999)
      rolling_count = 0;
    sprintf(tmp_buf, "%.3d", rolling_count);
    count_fp = fopen(COUNT_FILENAME, "w");
    fwrite(tmp_buf, 1, strlen(tmp_buf), count_fp);
    fclose(count_fp);
  }
  else
  {
    /* create the .count file */
    new_umask = S_IXUSR | S_IXGRP | S_IXOTH;
    old_umask = umask(new_umask);
    count_fp = fopen(COUNT_FILENAME, "w");
    strcpy(tmp_buf, "000");
    fwrite(tmp_buf, 1, strlen(tmp_buf), count_fp);
    fclose(count_fp);
    umask(old_umask);
    rolling_count = 0;
  }

  return(rolling_count);
}


#if defined(DEBUG)

int debug = FALSE;

#define DIRECTION_SEND  '>'
#define DIRECTION_RECV  '<'

Dump(
    char direction,
    unsigned char *buffer,
    int length)
{
  int i, j, numrows;
  unsigned char c;
/*char tc;*/

  numrows = (length + 15) / 16;
  for (i = 0; i < numrows; i++)
    {
      fprintf(stderr, "%c 0x%x\t", direction, i*16);
      fprintf(stderr, "    ");
      for (j = 0; j < 16; j++)
        {
          if (j + i * 16 >= length) fprintf(stderr, "  ");
          else fprintf(stderr, "%02x", buffer[j + i * 16] & 0xff);
          if (j % 4 == 3) putc(' ', stderr);
        }
      fprintf(stderr, " |");
      for (j = 0; j < 16; j++)
        {
          if (j + i * 16 >= length) putc(' ', stderr);
          else
            {
              c = buffer[j + i * 16];
              if (isprint(c)) {
                 putc(c, stderr);
              } else {
                 putc('.', stderr);
              } /* endif */
            }
        }
      fprintf(stderr, "|\n");
    }
}

int DebugSend( int socket, char *buffer, int len, int flags)
{
   int rc = send(socket, buffer, len, flags);

   fprintf(stderr, "---- send(%d, %p, %d, %x): rc=%d(0x%x)\n", socket, buffer, len, flags, rc, rc);
   Dump(DIRECTION_SEND, buffer, len);
   return(rc);
}

int DebugRecv( int socket, char *buffer, int len, int flags)
{
   int rc = recv(socket, buffer, len, flags);

   fprintf(stderr, "---- recv(%d, %p, %d, %x): rc=%d(0x%x)\n", socket, buffer, len, flags, rc, rc);
   if(rc >= 0) Dump(DIRECTION_RECV, buffer, rc);
   return(rc);
}

#undef send
#undef recv
#define send(_x,_y,_z,_w)   (debug?DebugSend(_x,_y,_z,_w):write(_x,_y,_z))
#define recv(_x,_y,_z,_w)   (debug?DebugRecv(_x,_y,_z,_w):read(_x,_y,_z))
#endif


#if defined(LPRMON)


ULONG SendEntryData(LQE *entry, ULONG count, int lprsock, int quietflag)
{
   static LISTENTRY *temp;
   static int   i;

        i      = 0;
        temp   = entry->List->Next;
        entry->SizePrinted = 0L;
        while( count > 0 ){
                if( (i = send( lprsock, temp->Data+i, count > MAXDATASIZE ? MAXDATASIZE : count, 0 )) <= 0 ) {
/*                 printf(cant_send);  */
                   soclose(lprsock);
                   return(0);
                }
                entry->SizePrinted += (ULONG) i;
                count  -= (ULONG) i;
                if( !quietflag ){
                   if( entry->SizePrinted > (ULONG) 1024 )
                      printf( "Sent %ld Kbytes.\r", entry->SizePrinted >> 10 );
                }
                if (i == MAXDATASIZE) {
                   i = 0;
                   NEXT(temp);
                } /* endif */
                UpdateStatus();
        }
        return(entry->SizePrinted);
}
#endif

#if defined(LPR)
ULONG FileSize(CHAR *file)
{
   int    i;
   long   fpos = 0L;
   BYTE   *filebuf;
   FILE   *fin;
#define FILESIZEBLOCK   2048
   if(filebuf = (BYTE *)malloc(FILESIZEBLOCK)) {
      if(( fin = fopen(file,"rt")) == (FILE *) NULL ) {
              fprintf(stderr, "Can't open %s: \n",file);
              return(0L);
      }

      while ((i=read(fileno(fin), filebuf, FILESIZEBLOCK))>0) {
         fpos += i;
      } /* endwhile */
      fclose(fin);
      free(filebuf);
   } else {
      perror("Unable to determine file size\n");
   } /* endif */
   return(fpos);
}

ULONG SendFileData(CHAR *file, ULONG count, int lprsock, int quietflag)
{
   static int    i;
   static ULONG  kbytes;
   static BYTE filebuf[2048];
   static FILE *fin;

        kbytes = 0L;
        if(( fin = fopen(file,bflag ? "rb" : "rt")) == (FILE *) NULL ) {
                fprintf(stderr, "Can't open %s: \n",file);
                return(0L);
        }

        while( count ){
                i = read( fileno( fin ), filebuf, sizeof( filebuf ));
                if( i < 0 ) {
                        fprintf(stderr, "Early EOF found transmitting file.\n" );
                        fclose(fin);
                        return(0L);
                }
                count -= (long) i;
#if 0
                if(count!=0) {
                      count = 0;
                      i--;
                }
#endif
                kbytes += (long) i;
                if( send( lprsock, filebuf, i, 0 ) != i ) {
                        fprintf(stderr, cant_send );
                        fclose(fin);
                        return(0L);
                }
                if( !quietflag ){
                        if( kbytes > (long) 1024 )
                                printf( "Sent %ld Kbytes.\r", kbytes >> 10 );
                }
        }
        fclose(fin);
        return(kbytes);
}
#endif

int HostConnect(char *name, int _port, int *_asock)
{
   struct sockaddr_in  _in_sin;
   struct hostent      h, *_host = 0;
   char                *addr_list[2];
   int                 _socket;
   int                 rc;

   if (_asock == NULL) {
      _socket = socket(AF_INET, SOCK_STREAM, 0);      /* allocate a socket */
   } else {
      _socket = *_asock;
   } /* endif */
   _in_sin.sin_port        = _port;                /* setup the port    */
   _in_sin.sin_addr.s_addr = inet_addr(name);      /* get address       */

   if (_in_sin.sin_addr.s_addr != -1) {            /* if passed an ip address */
      _in_sin.sin_family = AF_INET;
      addr_list[0]   = (char *) &_in_sin.sin_addr;
      addr_list[1]   = NULL;
      h.h_addrtype   = AF_INET;
      h.h_length     = sizeof _in_sin.sin_addr;
      h.h_addr_list  = addr_list;
      _host          = &h;
   } else {                                        /* else passed a hostname  */
      if((_host = gethostbyname(name)) != NULL) {  /* get the nameserver addresses */
         _in_sin.sin_family = _host->h_addrtype;   /* setup the addrtype           */
      } else {
         if (_asock == NULL) soclose(_socket);
         return(-1 * tcperrno());
      } /* endif */
   } /* endif */

   do {
      bcopy(*(_host->h_addr_list), (caddr_t)&(_in_sin.sin_addr), _host->h_length);
      if((rc=connect((_socket), (struct sockaddr *)&(_in_sin), sizeof(_in_sin))) < 0) {
          _host->h_addr_list++;
          printf("Unable to connect to %s (errno=%d)!\n", inet_ntoa(_in_sin.sin_addr), tcperrno());
          soclose(_socket);
          _socket = socket(AF_INET, SOCK_STREAM, 0);      /* allocate a socket */
      } /* endif */
   } while ((rc<0)               &&  /* if error on connect and              */
            (_socket != -1)      &&  /* still have a valid socket and        */
            (_host)              &&  /* still have a valid host entry and    */
            (_host)->h_addr_list &&  /* there are still addresses to try and */
            *(_host->h_addr_list));  /* the address is not NULL, try again   */

   if ((rc == -1) && (_asock == NULL)) soclose(_socket);

   if (_asock != NULL) *_asock = _socket;

   return((rc==-1) ? (-1*tcperrno()) : _socket);
}


/*----------------------------------------------------------------------------*/
#if defined(LPRMON)
#define printf lprmon_printf
#endif

USHORT lpr(VOID *temp, CHAR *file, ULONG count)
{
  /*    static int    c, i, j, len;  */
        static int    i, j, len;
        static int    rc;
        static ULONG  kbytes;
        static char   rbuf[256];
        static char   *addr_list[2];
        static struct sockaddr_in laddr;
        static struct sockaddr_in me;
        static struct sockaddr_in him;
        static struct hostent h, *hp;
        static int    melen = sizeof( me );
        static struct servent *sp;            /* service spec for tcp/lpr */
        static struct timeval tv;
/*      static int    on = 1;  */
        int lprsock = -1;
/*      int noretriestoconnect;  */
        int retry;
        char lpstr[1000];

        sp = getservbyname("printer", "tcp");
        if (sp == 0) {
                fprintf(stderr, "printer: printer/tcp: unknown service\n");
                exit(1);
        }


        him.sin_family = AF_INET;
        him.sin_port   = sp->s_port;


        /* if -S not specified, figure out something to use for print server name */
        if( server != (char *) NULL ){
                him.sin_addr.s_addr = inet_addr(server);
                if (him.sin_addr.s_addr != -1) {
                        addr_list[0]   = (char *) &him.sin_addr;
                        addr_list[1]   = NULL;
                        h.h_addrtype   = AF_INET;
                        h.h_length     = sizeof him.sin_addr;
                        h.h_addr_list  = addr_list;
                        hp             = &h;
                }
                else {
                  if(( hp = gethostbyname( server )) == (struct hostent *) NULL ) {
                        printf("unknown host\n");
                        return(LPR_RETRY | LPR_UNKNOWN_HOST);
                  }
                  bcopy(hp->h_addr, (caddr_t)&him.sin_addr, hp->h_length);
                  him.sin_family = hp->h_addrtype;
                  server = strdup( hp->h_name );
                }
        }

        if( server == (char *) NULL ) {
           printf("Out of memory\n" );
           return(LPR_ABORT | LPR_OUT_OF_MEMORY);
        }


        for (j = 0; hp->h_addr_list[j]; j++) {

           memcpy((char *) &him.sin_addr, hp->h_addr_list[j], hp->h_length);

           /* Now, connect to the host */
           if( !quietflag )
                   printf("Trying LPD print server %s(%s), device %s.\n",
                         server, inet_ntoa(him.sin_addr), printer );

           for (retry=0; retry<2; retry++) {
              /* allocate a socket  */
              if(( lprsock = socket( AF_INET, SOCK_STREAM, 0)) == -1 ) {
                 printf("Unable to allocate socket (errno() = %d)\n", tcperrno());
                 return(LPR_RETRY | LPR_OUT_OF_SOCKETS);
              }

              /* generate a "reserved" port number (< 1024) */
              for(rc = -1,i=721; rc && (i < 732) ;i++) { 
                 laddr.sin_addr.s_addr = INADDR_ANY;
                 laddr.sin_family      = AF_INET;
                 laddr.sin_port        = htons(i);
                 rc = bind(lprsock, (struct sockaddr *)&laddr, sizeof(laddr));
              }

              if(rc) {
                 printf("Unable to bind socket.\n");
                 soclose(lprsock);
                 return(LPR_RETRY | LPR_BIND);
              }

              if (connect(lprsock, (struct sockaddr *) &him, sizeof him) >= 0)
                break;

              soclose(lprsock);
              lprsock = -1;
           } /* endfor */

           if (lprsock != -1) break;

        } /* endfor */

        if (lprsock == -1) {
          printf("Unable to connect to foreigh host\n");
          return(LPR_RETRY | LPR_CONNECT);
        } /* endif */

        /*
        ** Connection is open, indicate we want to queue a file to the
        ** specified printer.
        */

        sprintf( buffer, "\2%s\n", printer );
        len = strlen( buffer );
        if( (send( lprsock, buffer, len, 0 ) != len )    ||
            recv_reply( lprsock, rbuf, sizeof( rbuf )))   {
               printf(cant_send);
               soclose(lprsock);
               return(LPR_ABORT | LPR_SEND);
        }

        /*
        ** Figure out the remote filename to use, sans first character.  The
        ** number field increases based on a rolling count file that is created
        ** in /tmp/.lprafp_count (000 - 999).
        ** The host name is our internet address in dotted decimal notation.
        */

        if( getsockname( lprsock, (struct sockaddr *) &me, &melen ) == -1 ) {
                printf("Unable to determine local address: \n");
                soclose(lprsock);
                return(LPR_ABORT | LPR_GERROR);
        }

        sprintf( filename, "fs%03d%s",get_rolling_count(),
                            inet_ntoa(me.sin_addr ));

        /*
        ** Figure out how long the file is.  This requires some hackery because
        ** of silly optional end-of-file characters. If there's a control-Z on
        ** the end we don't want to send it.
        */

        sprintf( buffer, "\3%lu d", count );
        strcat(buffer,filename);
        strcat(buffer,"\n");

        /* Send the file length and the filename */

        len = strlen( buffer );
        if( (send( lprsock, buffer, len, 0 ) != len ) ||
            recv_reply( lprsock, rbuf, sizeof( rbuf ))) {
                printf(cant_send);
                soclose(lprsock);
                return(LPR_ABORT | LPR_SEND);
        }

        /* Send the file */

        kbytes = (ULONG) 0;
        i = 0;
#if defined(LPRMON)
           if (!(kbytes = SendEntryData((LQE *)temp, count, lprsock, quietflag))) {
              return(LPR_ABORT);
           }
#endif

#if defined(LPR)
           if (!(kbytes = SendFileData(file, count, lprsock, quietflag))) {
              return(LPR_ABORT);
           }
#endif

        /*************************************************************/
        /* From this point in, when ever we use 'file' it is being   */
        /* Send within a command (ie. not as the file name of the    */
        /* local file to open and send).  Therefor, we need to       */
        /* to make sure that the file name will be interpereted      */
        /* by the server as a single file name.  Since SP is a       */
        /* significant character to the server, all spaces need      */
        /* to be converted (will use '_').                           */
        /*************************************************************/

        if ((cp = file) != NULL) {
           while (cp=strchr(cp,' ')) {
              *cp++ = '_';
           } /* endwhile */
        } else {

        } /* endif */

        if( !quietflag)
                printf( "Sent %ld bytes.\n", kbytes );

        /* Send '\0' indicating file is OK and close the file */

        if( (send( lprsock, "", 1, 0 ) != 1 )  ||
            recv_reply( lprsock, rbuf, sizeof( rbuf ))) {
                printf(cant_send);
                soclose(lprsock);
                return(LPR_ABORT | LPR_SEND);
        }

        /* Build the control file */

        hp = gethostbyaddr((char *)&(me.sin_addr), sizeof(struct in_addr), AF_INET);

        sprintf(cmds, "H%s\nP%s\nJ%s\nC%s\nL%s\n",
/*      sprintf(cmds, "H%s\nP%s\nJ%s\nC%s\nO%s\n", */
                      (hp ? hp->h_name : inet_ntoa(me.sin_addr)),
                      user, file, inet_ntoa(me.sin_addr), user);

        if( fflag ) {         /* Text file to be formatted through PR   */
           /* sprintf( tempcmds, "T%s\npd%s\n", file, filename); */
            sprintf( tempcmds, "pd%s\nT%s\n", filename, file ); 
           strcat(cmds, tempcmds);
        } else {
           if (bflag) {       /*  Binary file to be printed */
              sprintf( tempcmds, "ld%s\n", filename);
           } else {           /*  Text file to be printed   */
              sprintf( tempcmds, "fd%s\n", filename);
           } /* endif */
           strcat(cmds, tempcmds);
        }

/*      sprintf( tempcmds, "Ud%s\nN%s\n", filename, file ); */
        sprintf( tempcmds, "Ud%s\nN%s\n", filename, givenfilename ? givenfilename : file );
        strcat(cmds, tempcmds);

        strcat(cmds, lpdopts);

        sprintf( buffer, "\2%d c%s\n", strlen(cmds), filename );
        len = strlen( buffer );
        if( (send( lprsock, buffer, len, 0 ) != len )  ||
            recv_reply( lprsock, rbuf, sizeof( rbuf ))) {
                printf(cant_send);
                soclose(lprsock);
                return(LPR_ABORT | LPR_SEND);
        }

        len = strlen( cmds );
        if( (send( lprsock, cmds, len, 0 ) != len )    ||
            (send( lprsock, "", 1, 0 ) != 1 )          ||
            recv_reply( lprsock, rbuf, sizeof( rbuf )) ||
            (send( lprsock, "", 1, 0 ) != 1 )) {
                printf(cant_send);
                soclose(lprsock);
                return(LPR_ABORT | LPR_SEND);
        }

        if( !quietflag )
                printf( "The entire document was sent.\n" );

        soclose(lprsock);

        return(LPR_SUCCESS);
}

/*------------------------------------------------------------------------*/

static USHORT recv_reply(
int s,
char *buf,
unsigned len)
{
        register int i;
     /* int j; */
        int oktoprint = TRUE;

        if(( i = recv( s, buf, len, 0 )) == -1 ) {
                perror("recv()");
                printf("Unable to receive response from server!\n");
                return(1);
        }
        else if( i == 0 ) {
                printf("Server closed connection prematurely\n" );
                return(1);
        }
        else if( *buf == '\0' )
                return(0);
        else
                len = i;


        if( len > 1 ){
                for( i = 1; i < len && buf[ i ] != '\0'; i++ ){
                        if( isprint( buf[ i ] )) continue;
                        if( isspace( buf[ i ] )) continue;

                        oktoprint = FALSE;
                        break;
                }
        } else
                oktoprint = FALSE;

        printf( "Server Error" );

        switch( *buf ){
                case '\1':
                        printf( ": Server cannot open or write to printer" );
                        buf++;
                        break;

                case '\2':
                        printf( ": Out of storage space" );
                        buf++;
                        break;

                default:
                        if( !isprint( *buf )) oktoprint = FALSE;
                        if( oktoprint ) break;
                        printf( ": Unknown error: %d\n", *buf );
                        return(1);
        }

        if( oktoprint ){
                if( buf[ len - 1 ] != '\0' ) buf[ len - 1 ] = '\0';
                printf( ": %s", buf );
        }
        if( ( !oktoprint ) || buf[ len - 1 ] != '\n' )
                printf( "\n" );
/*      return(0);  */
        return(1);
}
#if defined(LPRMON)
#undef printf
#endif

/*----------------------------------------------------------------------------*/

void usage()
{
        printf( "\n" );
#if defined(LPRMON)
        printf( "Usage: \n"       \
                "  lprmon -?\n" \
                "  lprmon [-{f|b}] [-r<retries>] [-q<delay>] \n" \
                "         [-p<printer>] [-s<server>] <device>\n");
        printf( "\n\tdevice  - Parallel port to monitor" );
        printf( "\n\t          User should specify lpt<n>");
        printf( "\n\t          Where n = 1..m\n");
#endif

#if defined(LPR)
        printf( "Usage: \n"
                "  lprafp -?\n"
/*****************************************************************************/
/*              "  lprafp [-{f|b}] [-n] [-r<retries>] [-q<delay>] \n"        */
/*****************************************************************************/
                "  lprafp [-n] [-o<option>] [-r<retries>] [-q<delay>] \n" 
                "         [-p<printer>] [-s<server>] <files>\n");
        printf( "\n\tfiles   - Files to be sent to the server" );
        printf( "\n\t          Global file characters are accepted\n");
#endif
        printf( "\n\tOptions -\n" );
        printf( "\t\t-?          Display this message\n" );
/*****************************************************************************/
/*      printf( "\t\t-b       Send as Binary File to the server\n" );        */
/*      printf( "\t\t-f       Notify server to format output through PR\n" );*/
/*****************************************************************************/
#if defined(LPR)
        printf( "\t\t-n          Display Error messages only\n" );
#endif
#if defined(LPRMON)
        printf( "\t\t-n          Disable beep when errors displayed\n" );
#endif
        printf( "\t\t-o<option>  used to specify PSF/6000 options\n");
        printf( "\t\t            (example:  -oformdef=F1SAMPLE)\n");
        printf( "\t\t-r<retries> Sets the number of retries\n" );
        printf( "\t\t            Default = %d retries\n", retries);
        printf( "\t\t-q<delay>   Sets retry delay in seconds\n" );
        printf( "\t\t            Default = %d seconds\n", delay / 1000);
        printf( "\t\t-p<printer> Specifies Printer device name on the server\n" );
        printf( "\t\t            Defaults to the LPR_PRINTER environment variable\n");
        printf( "\t\t-s<server>  Specifies Printer server\n" );
        printf( "\t\t            Defaults to the LPR_SERVER environment variable\n");
}

/*********************************************************************/

#if defined(LPRMON)


/************************************************/
/* Routine to support calling of lpr from       */
/* multiple threads                             */
/************************************************/
USHORT  Protected_Lpr(LQE *entry)
{
   USHORT temp;
   static ULONG  LprSem=0L;

   REQUEST(&LprSem);
   temp = lpr(entry,entry->JobName,entry->Size);
   CLEAR(&LprSem);
   return(temp);
}
#endif


#if defined(LPR)

char *MakeMeATempFile(FILE *in,char *FileName)
{
 char TempBuffer[256];
 int  count;
 FILE *out;

 if(!(FileName = (char *)mktemp(FileName))) {
    fprintf(stderr, "Error: Unable to create temporary file!\n");
    return(NULL);
 } /* endif */

 if(!(out = fopen(FileName, "wb"))) {
    fprintf(stderr, "Error: Unable to open temporary file!\n");
    return(NULL);
 } /* endif */

 while ((count = fread(TempBuffer, 1, sizeof(TempBuffer), in)) > 0) {
    if(fwrite(TempBuffer, 1, count, out) != count) {
       fprintf(stderr, "Error: Unable to write to temporary file!\n");
       fclose(out);
       unlink(FileName);
       return(NULL);
    } /* endif */
 } /* endwhile */

 fclose(out);
 return(FileName);

}

main(
   int argc,
   char *argv[],
   char *envp[])
{
   CHAR   files[256];
   CHAR   *temp;
   int    i;
   USHORT rc;
   long   f_size;
   char   TempFileName[] = "LPXXXXXX";
   int    ReadFromStdin = 0;



   getopts(argv, files, lpdopts);
   temp = strtok(files, ";");
   if (!temp) {
      ReadFromStdin = 1;
      temp = MakeMeATempFile(stdin,TempFileName);
   } /* endif */
   while (temp != NULL) {

      f_size = FileSize(temp);
      if (f_size == 0L)
      {
        fprintf(stderr, "File does not exist or is zero bytes in length: %s\n", temp);
      }
      else
      {
        i  = 0;     /* reset retry count */
        if (!quietflag) printf("Printing %s:\n",temp);
        rc = lpr(NULL, temp, f_size);
        while (  ((rc & LPR_RETURN_MASK)==LPR_RETRY_FOREVER)  ||
                (((rc & LPR_RETURN_MASK)==LPR_RETRY) && (i++<(int)retries)) ) {
              sleep(delay/1000);
              printf("Printing(retry=%d) %s:\n", i, temp);
              rc=lpr(NULL, temp, f_size);
              switch (rc & LPR_RETURN_MASK) {
              case LPR_SUCCESS:
                 break;
              case LPR_ABORT:
                 fprintf(stderr, "Unable to send file %s\n", temp);
                 break;
              case LPR_RETRY:
                 if (i >= (int)retries) {
                    fprintf(stderr, "Unable to send file %s after %d tries!\n", temp, i);
                 } /* endif */
                 break;
              case LPR_RETRY_FOREVER:
                 break;
              default:;
              } /* endswitch */
        } /* endwhile */
      }
      if ( ReadFromStdin ) {
         unlink(temp);
         temp = NULL;
      } else {
         temp = strtok(NULL, ";");
      } /* endif */
   } /* endwhile */
}
#endif
